from scipy.io import wavfile
from pyreaper import reaper

from os.path import dirname, join
try:
    from tqdm import tqdm
except:
    def tqdm(x): return x


def test_reaper():
    N = 10
    for n in tqdm(range(N)):
        fs, x = wavfile.read(join(dirname(__file__), "test16k.wav"))
        pm_times, pm, f0_times, f0, corr = reaper(x, fs)



# coding: utf-8
# cython: boundscheck=True, wraparound=True

import numpy as np

cimport numpy as np
np.import_array()

cimport cython

from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector

from libcpp.cast cimport reinterpret_cast
from libc.stdint cimport int16_t, int32_t

from reaper cimport reaper as _reaper

cdef class Track:
    cdef _reaper.Track * ptr

    def __cinit__(self):
        self.ptr = new _reaper.Track()

    def __dealloc__(self):
        del self.ptr

    def num_frames(self):
        return self.ptr.num_frames()

    def num_channels(self):
        return self.num_channels()

    def set_v(self, int f, bool value):
        self.set_v(f, value)

    def resize(self, int n, int c=1):
        self.ptr.resize(n, c)


cdef class EpochTracker:
    """EpochTracker


    .. note::

        This isn't intended to be used directly in general. Use wrapper function instead
        unless if you don't want to understand internals.
    """
    cdef _reaper.EpochTracker * ptr

    def __cinit__(self):
        self.ptr = new _reaper.EpochTracker()

    def __dealloc__(self):
        del self.ptr

    def Init(self, np.ndarray[np.int16_t, ndim=1, mode="c"] input,
             float sample_rate, float min_f0_search, float max_f0_search,
             bool do_highpass, bool do_hilbert_transform):
        cdef int32_t n_input = len(input)
        return self.ptr.Init( & input[0], n_input, sample_rate, min_f0_search,
                             max_f0_search, do_highpass, do_hilbert_transform)

    def CleanUp(self):
        self.ptr.CleanUp()

    def set_debug_name(self, string name):
        self.ptr.set_debug_name(name)

    def debug_name(self):
        return self.ptr.debug_name()

    def ComputeFeatures(self):
        return self.ptr.ComputeFeatures()

    def TrackEpochs(self):
        return self.ptr.TrackEpochs()

    def set_unvoiced_cost(self, float v):
        self.ptr.set_unvoiced_cost(v)

    cdef ResampleAndReturnResults(self, float resample_interval,
                                  vector[float] * f0,
                                  vector[float] * correlations):
        return self.ptr.ResampleAndReturnResults(resample_interval, f0,
                                                 correlations)

    cdef GetFilledEpochs(self, float unvoiced_pm_interval,
                         vector[float] * times,
                         vector[int16_t] * voicing):
        self.ptr.GetFilledEpochs(unvoiced_pm_interval, times, voicing)

    # Convenient wrapper methods
    cdef GetEpochTrack(self, float inter_pulse, _reaper.Track * pm_track):
        cdef vector[float] times
        cdef vector[int16_t] voicing

        self.ptr.GetFilledEpochs(inter_pulse, & times, & voicing)

        pm_track.resize(times.size(), 1)

        cdef int32_t i
        for i in range(0, times.size()):
            ( & pm_track.t(i))[0] = times[i]
            pm_track.set_v(i, voicing[i])

        return True

    cdef GetTrackVoicedFlags(self, _reaper.Track * track,
                             np.ndarray[np.int32_t, ndim = 1, mode = "c"] voiced_flags):
        cdef int32_t i
        for i in range(0, track.num_frames()):
            voiced_flags[i] = 1 if track.v(i) else 0

    cdef GetTrackTimes(self, _reaper.Track * track, float * times):
        cdef int32_t i
        for i in range(0, track.num_frames()):
            times[i] = track.t(i)

    cdef GetTrackValues(self, _reaper.Track * track, float * values):
        cdef int32_t i
        for i in range(0, track.num_frames()):
            values[i] = track.a(i)

    cdef GetF0AndCorrTrack(self, float external_frame_interval,
                           _reaper.Track * f0_track,
                           _reaper.Track * corr_track):
        cdef vector[float] f0
        cdef vector[float] corr
        if not self.ResampleAndReturnResults(external_frame_interval,
                                             & f0, & corr):
            return False

        f0_track.resize(f0.size(), 1)
        corr_track.resize(corr.size(), 1)
        cdef int32_t i
        cdef float t
        for i in range(0, f0.size()):
            t = external_frame_interval * i
            ( & f0_track.t(i))[0] = t
            ( & corr_track.t(i))[0] = t
            f0_track.set_v(i, True if f0[i] > 0.0 else False)
            corr_track.set_v(i, True if f0[i] > 0.0 else False)
            (& f0_track.a(i))[0] = f0[i] if f0[i] > 0.0 else -1.0;
            (& corr_track.a(i))[0] = corr[i]

        return True


def reaper_internal(np.ndarray[np.int16_t, ndim=1, mode="c"] x, fs,
                    float minf0=40.0,
                    float maxf0=500.0,
                    bool do_high_pass=True,
                    bool do_hilbert_transform=False,
                    float inter_pulse=0.01,
                    float frame_period=0.005,
                    float unvoiced_cost=0.9):
    et = EpochTracker()
    et.set_unvoiced_cost(unvoiced_cost)
    ok = et.Init(x, fs, minf0, maxf0, do_high_pass, do_hilbert_transform)
    if not ok:
        raise RuntimeError("EpochTracker init failed")

    ok = et.ComputeFeatures()
    if not ok:
        raise RuntimeError("EpochTracker ComputeFeatures failed")

    ok = et.TrackEpochs()
    if not ok:
        raise RuntimeError("EpochTracker TrackEpochs failed")

    # Get pitch mark
    pm_track = Track()
    et.GetEpochTrack(inter_pulse, pm_track.ptr)

    cdef int pN = pm_track.num_frames()
    cdef np.ndarray[np.int32_t, ndim = 1, mode = "c"] pm \
        = np.zeros(pN, dtype=np.int32)
    cdef np.ndarray[np.float32_t, ndim= 1, mode = "c"] pm_times \
        = np.zeros(pN, dtype=np.float32)
    et.GetTrackTimes(pm_track.ptr, & pm_times[0])
    et.GetTrackVoicedFlags(pm_track.ptr, pm)

    # Get f0 and correlations
    f0_track = Track()
    corr_track = Track()
    ok = et.GetF0AndCorrTrack(frame_period, f0_track.ptr, corr_track.ptr)
    if not ok:
        raise RuntimeError("EpochTracker ResampleAndReturnResults failed")

    cdef int fN = f0_track.num_frames()
    assert fN == corr_track.num_frames()
    cdef np.ndarray[np.float32_t, ndim= 1, mode = "c"] f0 \
        = np.zeros(fN, dtype=np.float32)
    cdef np.ndarray[np.float32_t, ndim= 1, mode = "c"] corr \
        = np.zeros(fN, dtype=np.float32)
    cdef np.ndarray[np.float32_t, ndim= 1, mode = "c"] f0_times\
        = np.zeros(fN, dtype=np.float32)
    et.GetTrackTimes(f0_track.ptr, & f0_times[0])
    et.GetTrackValues(f0_track.ptr, & f0[0])
    et.GetTrackValues(corr_track.ptr, & corr[0])

    et.CleanUp()

    return pm_times, pm, f0_times, f0, corr



# coding: utf-8

"""
A python wrapper for `REAPER (Robust Epoch And Pitch EstimatoR)
<https://github.com/google/REAPER>`_.

https://github.com/r9y9/pyreaper
"""

from __future__ import division, print_function, absolute_import

import pkg_resources

__version__ = pkg_resources.get_distribution('pyreaper').version

from pyreaper.creaper import reaper_internal


def reaper(x, fs, minf0=40.0, maxf0=500.0, do_high_pass=True,
           do_hilbert_transform=False, inter_pulse=0.01,
           frame_period=0.005, unvoiced_cost=0.9):
    """REAPER (Robust Epoch And Pitch EstimatoR)

    Perform REAPER analysis given an audio signal

    Parameters
    ----------
    x : np.ndarray, dtype=np.int16
        Input audio signal

    fs : int
        Sampling frequency

    minf0 : float
        Min f0. Default is 40.0.

    maxf0 : float
        Max f0. Default is 500.0.

    do_high_pass : Bool
        Enable Rumble-removel highpass filter. Default is True.

    do_hilbert_transform : Bool
        Enable Hilbert transform that may reduce phase distortion.
        Default is False.

    inter_pulse : float
        Regular inter-mark interval to use in UV pitchmark regions.
        Default is 0.01 (sec)

    frame_period : float
        Frame period. Default is 0.005 (sec).

    unvoiced_cost : float
        Set the cost for unvoiced segments. Default is 0.9, the higher the value
        the more f0 estimates in noise.

    Returns
    -------
    pm_times : np.ndarray, dtype=np.float32
        Pitch mark time series in seconds

    pm : np.ndarray, dtype=np.int32
        Pitch mark. Value 1 and 0 means voiced frame and unvoiced frame,
        respectively.

    f0_times : np.ndarray, dtype=np.float32
        F0 time series in seconds

    f0 : np.ndarray, dtype=np.float32
        F0 contour

    corr : np.ndarray, dtype=np.float32
        Correlations

    Raises
    ------
    RuntimeError
        - if EpochTracker Init failed
        - if EpochTracker ComputeFeatures failed
        - if EpochTracker TrackEpochs failed
        - if EpochTracker ResampleAndReturnResults failed

    Examples
    --------
    >>> from scipy.io import wavfile
    >>> import pysptk
    >>> import pyreaper
    >>> fs, x = wavfile.read(pysptk.util.example_audio_file())
    >>> pm_times, pm, f0_times, f0, corr = pyreaper.reaper(x, fs)

    """
    return reaper_internal(x, fs, minf0, maxf0, do_high_pass,
                           do_hilbert_transform, inter_pulse, frame_period,
                           unvoiced_cost)



# distutils: language=c++

from libc.stdint cimport int16_t, int32_t
from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector

cdef extern from "core/track.h":
    cdef cppclass Track:
        Track()
        int num_frames() const
        int num_channels() const

        float & a(int f, int c=0)
        # float a(int f, int c=0) const
        float & t(int f)
        # float t(int f) const

        void set_v(int f, bool value)
        bool v(int f) const

        void resize(int n, int c)


cdef extern from "epoch_tracker/epoch_tracker.h":
    cdef cppclass EpochTracker:
        EpochTracker()
        void SetParameters()
        bool Init(const int16_t * input, int32_t n_input, float sample_rate,
                  float min_f0_search, float max_f0_search,
                  bool do_highpass, bool do_hilbert_transform)
        void set_debug_name(const string & debug_name)
        string debug_name()

        bool ComputeFeatures()
        bool TrackEpochs()

        bool ResampleAndReturnResults(float resample_interval,
                                      vector[float] * f0,
                                      vector[float] * correlations)

        void GetFilledEpochs(float unvoiced_pm_interval, vector[float] * times,
                             vector[int16_t] * voicing)

        void set_do_hilbert_transform(bool v)
        void set_do_highpass(bool v)
        void set_external_frame_interval(float v)
        void set_unvoiced_pulse_interval(float v)
        void set_min_f0_search(float v)
        void set_max_f0_search(float v)
        void set_unvoiced_cost(float v)

        void CleanUp()



# coding: utf-8

from __future__ import with_statement, print_function, absolute_import

from setuptools import setup, find_packages, Extension
from distutils.version import LooseVersion

import numpy as np
import os
from glob import glob
from os.path import join


min_cython_ver = '0.21.0'
try:
    import Cython
    ver = Cython.__version__
    _CYTHON_INSTALLED = ver >= LooseVersion(min_cython_ver)
except ImportError:
    _CYTHON_INSTALLED = False

try:
    if not _CYTHON_INSTALLED:
        raise ImportError('No supported version of Cython installed.')
    from Cython.Distutils import build_ext
    cython = True
except ImportError:
    cython = False

if cython:
    ext = '.pyx'
    cmdclass = {'build_ext': build_ext}
else:
    ext = '.cpp'
    cmdclass = {}
    if not os.path.exists(join("pyreaper", "creaper" + ext)):
        raise RuntimeError("Cython is required to generate C++ wrapper.")

# REAPER source location
src_top = join("lib", "REAPER")

src = glob(join(src_top, "core", "*.cc")) \
    + glob(join(src_top, "epoch_tracker", "*.cc"))
print(src)


# define core cython module
ext_modules = [Extension(
    name="pyreaper.creaper",
    sources=[join("pyreaper", "creaper" + ext)] + src,
    include_dirs=[np.get_include(), join(os.getcwd(), "lib", "REAPER")],
    extra_compile_args=[],
    language="c++",
)]

with open('README.md', 'r') as fd:
    long_description = fd.read()

setup(
    name='pyreaper',
    version='0.0.9',
    description='A python wrapper for REAPER (Robust Epoch And Pitch EstimatoR)',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Ryuichi Yamamoto',
    author_email='zryuichi@gmail.com',
    url='https://github.com/r9y9/pyreaper',
    license='MIT',
    packages=find_packages(),
    ext_modules=ext_modules,
    cmdclass=cmdclass,
    install_requires=[
        'numpy >= 1.8.0',
    ],
    tests_require=['nose', 'coverage'],
    extras_require={
        'test': ['nose', 'scipy'],
        'develop': ['cython >= ' + min_cython_ver],
    },
    classifiers=[
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Programming Language :: Cython",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
    ],
    keywords=["REAPER"]
)
